<?php

namespace Drupal\cl_block\Plugin\Block;

use Drupal\cl_components\Component\Component;
use Drupal\cl_components\Component\ComponentDiscovery;
use Drupal\cl_components\Exception\ComponentNotFoundException;
use Drupal\Component\Plugin\Exception\ContextException;
use Drupal\Component\Serialization\Json;
use Drupal\Component\Utility\NestedArray;
use Drupal\Core\Block\BlockBase;
use Drupal\Core\Cache\RefinableCacheableDependencyInterface;
use Drupal\Core\Form\FormStateInterface;
use Drupal\Core\Language\LanguageInterface;
use Drupal\Core\Plugin\ContainerFactoryPluginInterface;
use Drupal\Core\Render\BubbleableMetadata;
use Drupal\Core\StringTranslation\TranslatableMarkup;
use Drupal\Core\Utility\Token;
use Psr\Log\LoggerInterface;
use Shaper\Transformation\TransformationInterface;
use Shaper\Util\Context;
use Symfony\Component\DependencyInjection\ContainerInterface;

/**
 * Provides a block that renders a component.
 *
 * @Block(
 *   id = "cl_component_block",
 *   deriver = "\Drupal\cl_block\Plugin\Derivative\ComponentBlockDeriver",
 *   context_definitions = {
 *     "language" = @ContextDefinition("language", required = FALSE, label = @Translation("Language")),
 *     "node" = @ContextDefinition("entity:node", required = FALSE,
 *       label = @Translation("Node")),
 *     "user" = @ContextDefinition("entity:user", required = FALSE,
 *       label = @Translation("User Context"), constraints = { "NotNull" = {} },
 *     ),
 *   }
 * )
 *
 * @internal
 *   Plugin classes are internal.
 */
class ComponentBlock extends BlockBase implements ContainerFactoryPluginInterface {

  /**
   * The component discovery.
   *
   * @var \Drupal\cl_components\Component\ComponentDiscovery
   */
  protected ComponentDiscovery $componentDiscovery;

  /**
   * The field name.
   *
   * @var string
   */
  protected string $componentName;

  /**
   * The form generator from schema.
   *
   * @var \Shaper\Transformation\TransformationInterface
   */
  protected TransformationInterface $formGenerator;

  /**
   * The component
   *
   * @var \Drupal\cl_components\Component\Component
   */
  protected Component $component;

  /**
   * The logger.
   *
   * @var \Psr\Log\LoggerInterface
   */
  protected LoggerInterface $logger;

  /**
   * @var \Drupal\Core\Utility\Token
   */
  protected Token $token;

  /**
   * Does the site support inject.
   *
   * @var bool
   */
  protected bool $supportsInject = FALSE;

  /**
   * Constructs a new FieldBlock.
   *
   * @param array $configuration
   *   A configuration array containing information about the plugin instance.
   * @param string $plugin_id
   *   The plugin ID for the plugin instance.
   * @param mixed $plugin_definition
   *   The plugin implementation definition.
   * @param \Drupal\cl_components\Component\ComponentDiscovery $component_discovery
   *   The component discovery.
   * @param \Psr\Log\LoggerInterface $logger
   *   The logger.
   * @param \Shaper\Transformation\TransformationInterface $form_generator
   *   The form generator.
   * @param \Drupal\Core\Utility\Token $token
   *   The token replacement service.
   * @param bool $supports_inject
   *   TRUE if the site supports the inject syntax.
   */
  public function __construct(
    array $configuration,
    $plugin_id,
    $plugin_definition,
    ComponentDiscovery $component_discovery,
    LoggerInterface $logger,
    TransformationInterface $form_generator,
    Token $token,
    bool $supports_inject
  ) {
    $this->logger = $logger;
    $this->componentDiscovery = $component_discovery;
    $this->formGenerator = $form_generator;
    $this->token = $token;

    // Get the entity type and field name from the plugin ID.
    [, $component_name] = explode(static::DERIVATIVE_SEPARATOR, $plugin_id, 4);
    $this->componentName = $component_name;
    $this->supportsInject = $supports_inject;

    parent::__construct($configuration, $plugin_id, $plugin_definition);
  }

  /**
   * {@inheritdoc}
   */
  public static function create(ContainerInterface $container, array $configuration, $plugin_id, $plugin_definition) {
    return new static(
      $configuration,
      $plugin_id,
      $plugin_definition,
      $container->get(ComponentDiscovery::class),
      $container->get('logger.channel.cl_block'),
      $container->get('cl_block.form_generator'),
      $container->get('token'),
      $container->get('module_handler')->moduleExists('cl_inject')
    );
  }

  /**
   * {@inheritdoc}
   */
  public function build() {
    $config = $this->getConfiguration();
    $context = $config['component']['data'] ?? [];
    $twig_blocks = $config['component']['twig_blocks'] ?? [];
    $variant = $config['component']['variant'] ?? NULL;
    $component = $this->getComponent();
    $bubbleable_metadata = new BubbleableMetadata();
    $template = $this->generateComponentTemplate(
      $component,
      $variant,
      $twig_blocks,
      $context,
      $bubbleable_metadata
    );
    $build = [
      '#type' => 'inline_template',
      '#template' => $template,
      '#context' => $context,
    ];
    $bubbleable_metadata->applyTo($build);
    return $build;
  }

  /**
   * {@inheritdoc}
   */
  public function getPreviewFallbackString() {
    try {
      $component = $this->getComponent();
    }
    catch (ComponentNotFoundException $e) {
      return '';
    }
    $metadata = $component->getMetadata();
    return new TranslatableMarkup(
      '"@component" CL Component',
      ['@component' => $metadata->getName()]
    );
  }

  /**
   * {@inheritdoc}
   */
  public function defaultConfiguration() {
    return [
      'label_display' => FALSE,
      'component' => [
        'label' => 'above',
        'variant' => '',
        'data' => [],
        'twig_blocks' => [],
      ],
    ];
  }

  /**
   * {@inheritdoc}
   */
  public function blockForm($form, FormStateInterface $form_state) {
    $config = $this->getConfiguration();

    $form['component'] = [
      '#tree' => TRUE,
      '#process' => [
        [$this, 'componentDataProcessCallback'],
      ],
    ];
    $form['component']['label'] = [
      '#type' => 'select',
      '#title' => $this->t('Label'),
      '#options' => [
        'above' => $this->t('Above'),
        'inline' => $this->t('Inline'),
        'hidden' => '- ' . $this->t('Hidden') . ' -',
        'visually_hidden' => '- ' . $this->t('Visually Hidden') . ' -',
      ],
      '#default_value' => $config['component']['label'],
    ];

    $form['component']['variant'] = [
      '#type' => 'radios',
      '#title' => $this->t('Variant'),
      '#options' => $this->getComponentVariantOptions(),
      '#default_value' => $config['component']['variant'] ?? '',
      '#ajax' => [
        'callback' => [static::class, 'twigBlockAjaxCallback'],
        'wrapper' => 'twig-blocks-wrapper',
        'effect' => 'fade',
      ],
    ];

    try {
      $component = $this->getComponent();
    }
    catch (ComponentNotFoundException $e) {
      return $form;
    }
    // Here is where we derive the form from the metadata.
    $metadata = $component->getMetadata();
    $schemas = $metadata->getSchemas();
    $component_schema = $schemas['props'] ?? [];
    // Encode & decode, so we transform an associative array to an stdClass
    // recursively.
    $schema = json_decode(json_encode($component_schema));
    $form['component']['data'] = $form['component']['data'] ?? [];
    $element = &$form['component']['data'];
    $configuration = $this->getConfiguration();
    $current_input = $configuration['component']['data'] ?? [];
    $context = new Context([
      'form_state' => $form_state,
      'current_input' => $current_input,
      'form' => $form,
    ]);
    $element = $this->formGenerator->transform($schema, $context);
    $element['#weight'] = 2;
    // Add the component data to the form via AJAX.
    $form['component']['twig_blocks_wrapper'] = [
      '#prefix' => '<div id="twig-blocks-wrapper">',
      '#suffix' => '</div>',
      '#weight' => 5,
    ];
    $form['tree'] = [
      '#theme' => 'token_tree_link',
      '#token_types' => ['node', 'user'],
    ];
    return $form;
  }

  /**
   * Render API callback: builds the component data elements.
   */
  public function componentDataProcessCallback(array &$element, FormStateInterface $form_state): array {
    $component = NULL;
    try {
      $component = $this->getComponent();
    }
    catch (ComponentNotFoundException $e) {
      watchdog_exception('cl_block', $e);
    }
    if ($component) {
      $variant = $this->getVariantFromForm([
        ...$element['#parents'],
        'variant',
      ], $form_state);
      // Get the Twig block names, so we can make them configurable.
      $metadata = $component->getMetadata();
      $component_schema = $metadata->getSchemas()['named_blocks'][$variant] ?? [];
      if (empty($component_schema)) {
        return $element;
      }
      $twig_blocks = &$element['twig_blocks_wrapper']['twig_blocks'];
      $current_data = $this->getConfiguration()['component']['twig_blocks'] ?? [];
      // Process all block names.
      $block_names = array_keys($component_schema['properties'] ?? []);
      foreach ($block_names as $block_name) {
        $current_format = $current_data[$block_name]['format'] ?? 'full_html_with_twig';
        $current_value = $current_data[$block_name]['value'] ?? NULL;
        $twig_blocks[$block_name] = [
          '#type' => 'text_format',
          '#title' => $this->t(
            'Twig Block: @name',
            ['@name' => $block_name]
          ),
          '#format' => $current_format,
          '#default_value' => $current_value,
        ];
      }
      $twig_blocks['#parents'] = array_merge(
        $element['#parents'],
        ['twig_blocks']
      );
      // Store the array parents for our element so that we can retrieve the
      // twig block settings in our AJAX callback.
      $form_state->set('twig_block_array_parents', $element['#array_parents']);
    }
    $element['#weight'] = 5;
    return $element;
  }

  /**
   * Render API callback: gets the layout settings elements.
   */
  public static function twigBlockAjaxCallback(array $form, FormStateInterface $form_state) {
    $twig_block_array_parents = $form_state->get('twig_block_array_parents');
    return NestedArray::getValue($form, array_merge($twig_block_array_parents, ['twig_blocks_wrapper']));
  }

  /**
   * {@inheritdoc}
   */
  public function blockSubmit($form, FormStateInterface $form_state) {
    $this->configuration['component'] = $form_state->getValue('component');
  }

  /**
   * Get the available options for the variant selector.
   *
   * @return array
   *   The variant options.
   */
  protected function getComponentVariantOptions(): array {
    try {
      $component = $this->getComponent();
    }
    catch (ComponentNotFoundException $e) {
      return [];
    }
    $variants = $component->getVariants();
    $toHuman = fn(string $variant) => ucfirst(strtr($variant, '-', ' '));
    return [
      '' => $this->t('-Default variant-'),
      ...array_combine($variants, array_map($toHuman, $variants)),
    ];
  }

  /**
   * Gets the component.
   *
   * @return \Drupal\cl_components\Component\Component
   *   The component.
   *
   * @throws \Drupal\cl_components\Exception\ComponentNotFoundException
   */
  protected function getComponent(): Component {
    if (empty($this->component)) {
      $this->component = $this->componentDiscovery->find($this->componentName);
    }
    return $this->component;
  }

  /**
   * Gets the formatter object.
   *
   * @param array $parents
   *   The #parents of the element representing the formatter.
   * @param \Drupal\Core\Form\FormStateInterface $form_state
   *   The current state of the form.
   *
   * @return string
   *   The variant.
   */
  protected function getVariantFromForm(array $parents, FormStateInterface $form_state): string {
    // Use the processed values, if available.
    $variant = NestedArray::getValue($form_state->getValues(), $parents);
    if (isset($variant)) {
      return $variant;
    }
    // Next check the raw user input.
    $variant = NestedArray::getValue($form_state->getUserInput(), $parents);
    if (isset($variant)) {
      return $variant;
    }
    // If no user input exists, use the default values.
    return $this->getConfiguration()['component']['variant'] ?? '';
  }

  /**
   * Generates the template to render the component.
   *
   * @param string $component_id
   *   The component ID.
   * @param mixed $variant
   *   The variant.
   * @param array $twig_blocks
   *   The contents of any potential embed blocks.
   *
   * @return string
   *   The template.
   *
   * @throws \Drupal\cl_components\Exception\TemplateNotFoundException
   */
  private function generateComponentTemplate(Component $component, string $variant, array $twig_blocks, array $context, RefinableCacheableDependencyInterface $bubbleable_metadata): string {
    $token_data = array_filter($this->getContextValues());
    $metadata = $component->getMetadata();
    $component_path = $metadata->getPath();
    $template_name = $component->getTemplateName($variant);
    $template_path = $component_path . DIRECTORY_SEPARATOR . $template_name;
    $template = '{# This template was dynamically generated by cl_block #}' . PHP_EOL;
    // Replace possible tokens in the context mapping.
    $context = $this->replaceTokensRecursively($context, $token_data, $bubbleable_metadata);
    $template .= empty($context)
      ? sprintf('{%% embed \'%s\' %%}', $template_path)
      : sprintf('{%% embed \'%s\' with %s %%}', $template_path, Json::encode($context));
    $template .= PHP_EOL;
    $options = ['clear' => TRUE];
    try {
      // If there is a language in the context of the block, use it.
      $language = $this->getContextValue('language');
      if ($language instanceof LanguageInterface) {
        $options['langcode'] = $language->getId();
      }
    }
    catch (ContextException $e) {
      // Intentionally left blank.
    }
    foreach ($twig_blocks as $block_name => $block_value) {
      $bl_val = $this->token->replace(
        $block_value['value'],
        $token_data,
        $options,
        $bubbleable_metadata
      );
      $block_build = [
        '#type' => 'processed_text',
        '#text' => $bl_val,
        '#format' => is_array($block_value) ? $block_value['format'] : 'full_html_with_twig',
      ];
      try {
        $value = \Drupal::service('renderer')->render($block_build);
      }
      catch (\Exception $e) {
        $value = $bl_val;
      }
      $template .= "  {% block $block_name %}" . PHP_EOL
        . "    $value" . PHP_EOL
        . "  {% endblock %}" . PHP_EOL;
    }
    $template .= '{% endembed %}' . PHP_EOL;
    return $template;
  }

  /**
   * Replaces tokens recursivelly in a data structure.
   *
   * @param array $context
   *   The context.
   * @param array $token_data
   *   The token data.
   * @param \Drupal\Core\Cache\RefinableCacheableDependencyInterface $bubbleable_metadata
   *   The bubbleable metadata.
   *
   * @return array
   *   The context with the tokens replaced.
   */
  private function replaceTokensRecursively(array $context, array $token_data, RefinableCacheableDependencyInterface $bubbleable_metadata): array {
    return array_map(
      function (mixed $item) use ($token_data, $bubbleable_metadata): mixed {
        if (is_string($item)) {
          return $this->token->replacePlain(
            $item,
            $token_data,
            ['clear' => TRUE],
            $bubbleable_metadata
          );
        }
        if (is_array($item)) {
          return $this->replaceTokensRecursively($item, $token_data, $bubbleable_metadata);
        }
        return $item;
      },
      $context
    );
  }

}
